from collections.abc import Iterable
from os import remove, system
from os.path import isfile, join, split
from typing import TYPE_CHECKING, Any

from numpy import asarray

from matplotlib.pyplot import figure

from .xfoilpolar import XfoilPolar, write_polar_session
from .xfoilresult import XfoilResult, write_result_session

if TYPE_CHECKING:
    from matplotlib.axes import Axes
    from numpy.typing import NDArray


class Xfoil:
    name: str = None
    ppar: int = None
    x: 'NDArray' = None
    y: 'NDArray' = None
    results: dict[str, 'XfoilResult'] = None
    polars: dict[str, 'XfoilPolar'] = None
    frmstr: str = None
    _area: float = None

    def __init__(self, name: str, frmstr: str = '11.6f') -> None:
        self.name = name
        self.frmstr = frmstr

    def points_from_dat(self, datfile: str) -> None:
        x = []
        y = []
        with open(datfile, 'rt') as file:
            for i, line in enumerate(file):
                line = line.rstrip('\n')
                if i == 0:
                    self.name = line.strip()
                else:
                    split = line.split()
                    if len(split) == 2:
                        x.append(float(split[0]))
                        y.append(float(split[1]))
        self.x = asarray(x)
        self.y = asarray(y)
        self._area = None

    def set_points(self, x: Iterable[float], y: Iterable[float]) -> None:
        self.x = asarray(x)
        self.y = asarray(y)
        self._area = None

    def set_ppar(self, ppar: int) -> None:
        self.ppar = ppar

    @property
    def area(self) -> float:
        if self._area is None:
            xa, ya = self.x[:-1], self.y[:-1]
            xb, yb = self.x[1:], self.y[1:]
            areax2 = sum(xa*yb - xb*ya)
            xa, ya = self.x[-1], self.y[-1]
            xb, yb = self.x[0], self.y[0]
            areax2 += xa*yb - xb*ya
            self._area = 0.5*areax2
        return self._area

    def write_dat(self) -> str:

        from pyxfoil import workdir

        datname = self.name.replace(' ', '_')
        filepath = join(workdir, datname)
        datfilepath = f'{filepath:s}.dat'
        num = len(self.x)
        if self.area >= 0:
            order = range(num-1, -1, -1)
        else:
            order = range(num)
        frmstr = '  {:' + self.frmstr + '} {:' + self.frmstr + '}\n'
        with open(datfilepath, 'wt') as file:
            file.write(self.name+'\n')
            for i in order:
                file.write(frmstr.format(self.x[i], self.y[i]))
        return datfilepath

    def run_result(self, alfa: float, Re: float=None,
                   mach: float=None) -> 'XfoilResult':

        from pyxfoil import xfoilexe

        datfilepath = self.write_dat()
        numpnl = len(self.x) - 1
        sesfilepath, resfilepath = write_result_session(self.name, datfilepath, numpnl,
                                                        alfa, mach=mach, Re=Re,
                                                        ppar=self.ppar)

        if isfile(resfilepath):
            remove(resfilepath)

        if xfoilexe is None:
            err = 'Cannot locate "xfoil.exe" in path. '
            err += 'Import set_xfoilexe and use it to directly point to "xfoil.exe".'
            raise SystemError(err)

        system('{:s} < {:s}'.format(xfoilexe, sesfilepath))

        res = split(resfilepath)[1]
        res = res.replace('.res', '')
        result = XfoilResult(res, numpnl)
        result.set_param(alfa, mach, Re)
        result.read_result(resfilepath)
        if self.results is None:
            self.results = {}
        self.results[res] = result
        return result

    def run_polar(self, almin: float, almax: float, alint: float,
                  Re: float | None = None, mach: float | None = None,
                  xtrtop: float = 1.0, xtrbot: float = 1.0) -> 'XfoilPolar':

        from pyxfoil import xfoilexe

        datfilepath = self.write_dat()
        numpnl = len(self.x) - 1
        sesfilepath, polfilepath = write_polar_session(self.name, datfilepath,
                                                       numpnl, almin, almax, alint,
                                                       mach=mach, Re=Re, ppar=self.ppar,
                                                       xtrtop=xtrtop, xtrbot=xtrbot)

        if isfile(polfilepath):
            remove(polfilepath)

        if xfoilexe is None:
            err = 'Cannot locate "xfoil.exe" in path. '
            err += 'Import set_xfoilexe and use it to directly point to "xfoil.exe".'
            raise SystemError(err)

        system('{:s} < {:s}'.format(xfoilexe, sesfilepath))

        pol: str = split(polfilepath)[1]
        pol = pol.replace('.pol', '')
        polar = XfoilPolar(pol, numpnl)
        polar.read_polar(polfilepath)
        if self.polars is None:
            self.polars = {}
        self.polars[pol] = polar
        return polar

    def plot_profile(self, ax: 'Axes | None' = None, **kwargs: dict[str, Any]) -> 'Axes':
        if ax is None:
            figsize = kwargs.pop('figsize', (12, 8))
            grid = kwargs.pop('grid', True)
            aspect = kwargs.pop('aspect', 'equal')
            fig = figure(figsize=figsize)
            ax = fig.gca()
            ax.grid(grid)
            ax.set_aspect(aspect)
            ax.set_title('Airfoil profile for {:}.'.format(self.name))
        ax.plot(self.x, self.y, **kwargs)
        return ax

    def __repr__(self) -> str:
        return f'Xfoil({self.name:s})'

    def __str__(self) -> str:
        return f'Xfoil({self.name:s})'



from os import curdir


def find_xfoil(exe):
    def is_exe(fpath):
        from os import X_OK, access
        from os.path import isfile
        return isfile(fpath) and access(fpath, X_OK)
    from os import environ, pathsep
    from os.path import join, split
    fpath, _ = split(exe)
    if fpath:
        if is_exe(exe):
            return exe
    else:
        for path in environ["PATH"].split(pathsep):
            exe_file = join(path, exe)
            if is_exe(exe_file):
                return exe_file
    return None


xfoilexe = find_xfoil('xfoil.exe')
workdir = None


def set_xfoilexe(xfoil):
    global xfoilexe
    xfoilexe = xfoil


def set_workdir(wdir):
    from os.path import join
    wdir = join(wdir, '')
    if ' ' in wdir:
        print('The workding directory should have no spaces.')
    global workdir
    workdir = wdir


set_workdir(curdir)


from .xfoil import Xfoil as Xfoil



from os.path import join
from typing import TYPE_CHECKING, Any

from matplotlib.pyplot import figure
from numpy import asarray, radians

if TYPE_CHECKING:
    from matplotlib.axes import Axes
    from numpy.typing import NDArray


class XfoilPolar:
    name: str = None
    numpnl: int = None
    Re: float | None = None
    mach: float | None = None
    ncrit: float = None
    xtrftop: float = None
    xtrfbot: float = None
    alpha: 'NDArray' = None
    cl: 'NDArray' = None
    cd: 'NDArray' = None
    cdp: 'NDArray' = None
    cm: 'NDArray' = None
    trtop: 'NDArray' = None
    trbot: 'NDArray' = None
    _alrad: 'NDArray' = None
    _clocd: 'NDArray' = None

    def __init__(self, name: str, numpnl: int) -> None:
        self.name = name
        self.numpnl = numpnl

    def reset(self) -> None:
        for attr in self.__dict__:
            if attr.startswith('_'):
                setattr(self, attr, None)

    def read_polar(self, filepath: str) -> None:
        alpha = []
        cl = []
        cd = []
        cdp = []
        cm = []
        trtop = []
        trbot = []
        read = False
        with open(filepath, 'rt') as file:
            for line in file:
                line = line.rstrip('\n')
                if line.strip() == '':
                    continue
                if line[0:23] == ' Calculated polar for: ':
                    self.name = line[23:].strip()
                if line[0:8] == ' xtrf = ':
                    self.xtrftop = float(line[8:15])
                    self.xtrfbot = float(line[27:34])
                if line[0:8] == ' Mach = ':
                    self.mach = float(line[8:15])
                    self.Re = float(line[25:38].replace(' ', ''))
                    self.ncrit = float(line[51:58])
                if line[0:8] == '   alpha':
                    read = True
                    continue
                if read:
                    if line[0:8] != '  ------':
                        alpha.append(float(line[0:8]))
                        cl.append(float(line[8:17]))
                        cd.append(float(line[17:27]))
                        cdp.append(float(line[27:37]))
                        cm.append(float(line[37:46]))
                        trtop.append(float(line[46:55]))
                        trbot.append(float(line[55:64]))
        self.alpha = asarray(alpha)
        self.cl = asarray(cl)
        self.cd = asarray(cd)
        self.cdp = asarray(cdp)
        self.cm = asarray(cm)
        self.trtop = asarray(trtop)
        self.trbot = asarray(trbot)
        self.reset()

    @property
    def alrad(self) -> 'NDArray':
        if self._alrad is None:
            self._alrad = radians(self.alpha)
        return self._alrad

    @property
    def clocd(self) -> 'NDArray | None':
        if self._clocd is None:
            if self.cl is not None and self.cd is not None:
                clocdval = [cli/cdi for cli, cdi in zip(self.cl, self.cd)]
                self._clocd = asarray(clocdval)
        return self._clocd

    def plot_polar(self, xaxis='cd', yaxis='cl',
                   ax: 'Axes | None'=None, **kwargs: dict[str, Any]) -> 'Axes':
        if ax is None:
            figsize = kwargs.pop('figsize', (12, 8))
            fig = figure(figsize = figsize)
            ax = fig.gca()
            title = fr'Polar plot for $M = {self.mach:g}$ and $Re = {self.Re:.12g}$'
            ax.set_title(title.format(self.mach, self.Re))
            grid = kwargs.pop('grid', True)
            ax.grid(grid)
            xlabel = self.get_label(xaxis)
            ylabel = self.get_label(yaxis)
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)
        label = r''
        if self.Re is not None:
            label += r'$Re = {:.12g}$'.format(self.Re)
        if self.Re is not None and self.mach is not None:
            label += r'; '
        if self.mach is not None:
            label += r'$M = {:g}$'.format(self.mach)
        xvalue = self.get_value(xaxis)
        yvalue = self.get_value(yaxis)
        kwargs.setdefault('label', label)
        ax.plot(xvalue, yvalue, **kwargs)
        return ax

    def get_label(self, var: str) -> str:
        if var == 'alpha':
            label = r'$\alpha$'
        elif var == 'cl':
            label = '$C_l$'
        elif var == 'cd':
            label = '$C_d$'
        elif var == 'cdp':
            label = '$C_{dp}$'
        elif var == 'cm':
            label = '$C_m$'
        elif var == 'toptr':
            label = 'Top X Transition'
        elif var == 'bottr':
            label = 'Bottom X Transition'
        elif var == 'clocd':
            label = r'$\frac{c_l}{c_d}$'
        else:
            raise ValueError(f'{var:s} does not exist in XfoilPolar.')
        return label

    def get_value(self, var: str) -> 'NDArray':
        if var == 'alpha':
            value = self.alpha
        elif var == 'cl':
            value = self.cl
        elif var == 'cd':
            value = self.cd
        elif var == 'cdp':
            value = self.cdp
        elif var == 'cm':
            value = self.cm
        elif var == 'toptr':
            value = self.trtop
        elif var == 'bottr':
            value = self.trbot
        elif var == 'clocd':
            value = self.clocd
        else:
            raise ValueError(f'{var:s} does not exist in XfoilPolar.')
        return value

    def __repr__(self) -> str:
        return f'<pyxfoil.XfoilPolar {self.name:s}>'


def write_polar_session(name: str, datfilepath: str, numpnl: int,
                        almin: float, almax: float, alint: float,
                        mach: float | None = None,
                        Re: float | None = None,
                        ppar: int | None = None,
                        xtrtop: float = 1.0,
                        xtrbot: float = 1.0) -> tuple[str, str]:

    from pyxfoil import workdir

    polname = name.replace(' ', '_') + f'_{numpnl:d}'

    if mach is not None:
        polname += f'_{mach:g}'

    if Re is not None:
        polname += f'_{Re:.12g}'

    filepath = join(workdir, polname)
    sesfilepath = f'{filepath:s}.ses'
    polfilepath = f'{filepath:s}.pol'

    with open(sesfilepath, 'wt') as file:

        file.write('load {:s}\n'.format(datfilepath))

        if ppar is not None:
            file.write('ppar\n')
            file.write('n {:d}\n'.format(ppar))
            file.write('\n')
            file.write('\n')

        file.write('oper\n')

        if mach is not None:
            file.write('mach {:g}\n'.format(mach))

        if Re is not None:
            file.write('visc {:.12g}\n'.format(Re))

        # Set TRIP Position:
        file.write('VPAR\n')
        file.write('XTR\n')
        file.write(f'{xtrtop:g}\n')
        file.write(f'{xtrbot:g}\n')
        file.write('\n')

        file.write('pacc\n')
        file.write('{:s}\n'.format(polfilepath))
        file.write('\n')

        file.write('aseq\n')
        file.write('{:g}\n'.format(almin))
        file.write('{:g}\n'.format(almax))
        file.write('{:g}\n'.format(alint))

        file.write('pacc\n')

        if mach is not None:
            file.write('mach 0.0\n')

        if Re is not None:
            file.write('visc\n')

        file.write('\n')

        file.write('quit\n')

    return sesfilepath, polfilepath



from os.path import join
from typing import TYPE_CHECKING, Any

from matplotlib.pyplot import figure
from numpy import asarray, radians

if TYPE_CHECKING:
    from matplotlib.axes import Axes
    from numpy.typing import NDArray


class XfoilResult:
    name: str = None
    numpnl: int = None
    alpha: float = None
    Re: float | None = None
    mach: float | None = None
    s: 'NDArray' = None
    x: 'NDArray' = None
    y: 'NDArray' = None
    ue: 'NDArray' = None
    ds: 'NDArray' = None
    th: 'NDArray' = None
    cf: 'NDArray' = None
    h: 'NDArray' = None
    resfile: str = None
    _alrad: 'NDArray' = None
    _cp: 'NDArray' = None

    def __init__(self, name: str, numpnl: int) -> None:
        self.name = name
        self.numpnl = numpnl

    def reset(self) -> None:
        for attr in self.__dict__:
            if attr.startswith('_'):
                setattr(self, attr, None)

    def set_param(self, alpha: float, mach: float, Re: float) -> None:
        self.alpha = alpha
        self.mach = mach
        self.Re = Re

    def read_result(self, resfile: str) -> None:
        with open(resfile, 'rt') as f:
            s = []
            x = []
            y = []
            ue = []
            ds = []
            th = []
            cf = []
            h = []
            for line in f:
                line = line.rstrip('\n')
                if line != '':
                    if line[0] != '#':
                        s.append(float(line[1:11]))
                        x.append(float(line[11:20]))
                        y.append(float(line[20:29]))
                        ue.append(float(line[29:38]))
                        ds.append(float(line[38:48]))
                        th.append(float(line[48:58]))
                        cf.append(float(line[58:68]))
                        h.append(float(line[68:78]))
        self.s = asarray(s)
        self.x = asarray(x)
        self.y = asarray(y)
        self.ue = asarray(ue)
        self.ds = asarray(ds)
        self.th = asarray(th)
        self.cf = asarray(cf)
        self.h = asarray(h)
        self.reset()

    @property
    def alrad(self) -> 'NDArray':
        if self._alrad is None:
            self._alrad = radians(self.alpha)
        return self._alrad

    @property
    def cp(self) -> 'NDArray':
        if self._cp is None:
            self._cp = asarray([1-uei**2 for uei in self.ue])
        return self._cp

    def plot_result(self, xaxis='x', yaxis='ue', ax: 'Axes| None' = None,
                    **kwargs: dict[str, Any]) -> 'Axes':
        if ax is None:
            figsize = kwargs.pop('figsize', (12, 8))
            fig = figure(figsize=figsize)
            ax = fig.gca()
            grid = kwargs.pop('grid', True)
            ax.grid(grid)
            xlabel = self.get_label(xaxis)
            ylabel = self.get_label(yaxis)
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)
            title = r'Result plot for $\alpha = {:g}$'.format(self.alpha)
            if self.Re is not None:
                title += r' and $Re = {:.12g}$'.format(self.Re)
            if self.mach is not None:
                title += r' and $M = {:g}$'.format(self.mach)
            ax.set_title(title)
            if yaxis == 'cp':
                ax.invert_yaxis()
        label = r'$\alpha = {:g}$'.format(self.alpha)
        if self.Re is not None:
            label += r'; $Re = {:.12g}$'.format(self.Re)
        if self.mach is not None:
            label += r'; $M = {:g}$'.format(self.mach)
        xvalue = self.get_value(xaxis)
        yvalue = self.get_value(yaxis)
        kwargs.setdefault('label', label)
        ax.plot(xvalue, yvalue, **kwargs)
        return ax

    def result(self, var: str, correct: bool=False) -> 'NDArray':
        res = self.get_value(var)
        if correct:
            if var == 's':
                offset = max(res[:self.numpnl+1])
                val = [offset-resi for resi in res[:self.numpnl+1]]
            else:
                val = [resi for resi in res[:self.numpnl+1]]
            val.reverse()
            val = val + res[self.numpnl+1:]
        else:
            val = res.copy()
        return val

    def get_label(self, var: str) -> str:
        if var == 'x':
            label = '$x$'
        elif var == 'y':
            label = '$y$'
        elif var == 's':
            label = '$s$'
        elif var == 'ue':
            label = '$u_e$'
        elif var == 'cp':
            label = '$c_p$'
        elif var == 'ds':
            label = r'$\delta^*$'
        elif var == 'th':
            label = r'$\theta$'
        elif var == 'h':
            label = '$h$'
        elif var == 'cf':
            label = '$c_f$'
        else:
            raise ValueError(f'{var:s} does not exist in XfoilResult.')
        return label

    def get_value(self, var: str) -> 'NDArray':
        if var == 'x':
            value = self.x
        elif var == 'y':
            value = self.y
        elif var == 's':
            value = self.s
        elif var == 'ue':
            value = self.ue
        elif var == 'cp':
            value = self.cp
        elif var == 'ds':
            value = self.ds
        elif var == 'th':
            value = self.th
        elif var == 'h':
            value = self.h
        elif var == 'cf':
            value = self.cf
        else:
            raise ValueError(f'{var:s} does not exist in XfoilResult.')
        return value

    def __repr__(self) -> str:
        return f'<pyxfoil.XfoilResult {self.name:s}>'


def write_result_session(name: str, datfilepath: str, numpnl: int,
                         alpha: float, mach: float | None = None,
                         Re: float | None = None,
                         ppar: int | None = None) -> tuple[str, str]:

    from pyxfoil import workdir

    resname = name.replace(' ', '_')
    resname += f'_{numpnl:d}_{alpha:g}'
    if mach is not None:
        resname += f'_{mach:g}'
    if Re is not None:
        resname += f'_{Re:.12g}'
    filepath = join(workdir, resname)
    sesfilepath = f'{filepath:s}.ses'
    resfilepath = f'{filepath:s}.res'
    with open(sesfilepath, 'wt') as file:
        file.write('load {:s}\n'.format(datfilepath))
        if ppar is not None:
            file.write('ppar\n')
            file.write('n {:d}\n'.format(ppar))
            file.write('\n')
            file.write('\n')
        file.write('oper\n')
        if mach is not None:
            file.write('mach {:g}\n'.format(mach))
        if Re is not None:
            file.write('visc {:.12g}\n'.format(Re))
        file.write('alfa {:g}\n'.format(alpha))
        file.write('dump {:s}\n'.format(resfilepath))
        if mach is not None:
            file.write('mach 0.0\n')
        if Re is not None:
            file.write('visc\n')
        file.write('\n')
        file.write('quit\n')

    return sesfilepath, resfilepath
